@@ -141,10 +141,18 @@ terminal_receiver_impl_exec (TerminalReceiver *receiver,
     shell = FALSE;
   if (!g_variant_lookup (options, "environ", "^a&ay", &envv))
     envv = NULL;
-
   if (!g_variant_lookup (options, "fd-set", "@a(ih)", &fd_array))
     fd_array = NULL;
 
+  /* Check environment */
+  if (!terminal_util_check_envv((const char * const*)envv)) {
+    g_dbus_method_invocation_return_error_literal (invocation,
+                                                   G_DBUS_ERROR,
+                                                   G_DBUS_ERROR_INVALID_ARGS,
+                                                   "Malformed environment");
+    goto out;
+  }
+
   /* Check FD passing */
   if ((fd_list != NULL) ^ (fd_array != NULL)) {
     g_dbus_method_invocation_return_error_literal (invocation,
@@ -164,6 +172,13 @@ terminal_receiver_impl_exec (TerminalReceiver *receiver,
       const int fd = fd_array_data[2 * i];
       const int idx = fd_array_data[2 * i + 1];
 
+      if (fd == -1) {
+        g_dbus_method_invocation_return_error (invocation,
+                                               G_DBUS_ERROR,
+                                               G_DBUS_ERROR_INVALID_ARGS,
+                                               "Passing of invalid FD %d not supported", fd);
+        goto out;
+      }
       if (fd == STDIN_FILENO ||
           fd == STDOUT_FILENO ||
           fd == STDERR_FILENO) {
@@ -192,7 +207,13 @@ terminal_receiver_impl_exec (TerminalReceiver *receiver,
 
   ExecData *exec_data = g_new (ExecData, 1);
   exec_data->receiver = g_object_ref (receiver);
-  exec_data->invocation = invocation; /* adopted */
+  /* We want to transfer the ownership of @invocation to ExecData here, but
+   * we have to temporarily ref it so that in the error case below (where
+   * terminal_screen_exec() frees the exec data via the supplied callback,
+   * the g_dbus_method_invocation_take_error() calll still can take ownership
+   * of the invocation's ref passed to this function (terminal_receiver_impl_exec()).
+   */
+  exec_data->invocation = g_object_ref (invocation);
 
   GError *err = NULL;
   if (!terminal_screen_exec (priv->screen,
@@ -206,9 +227,13 @@ terminal_receiver_impl_exec (TerminalReceiver *receiver,
                              (GDestroyNotify) exec_data_free,
                              NULL /* cancellable */,
                              &err)) {
+    /* Transfers ownership of @invocation */
     g_dbus_method_invocation_take_error (invocation, err);
   }
 
+  /* Now we can remove that extra ref again. */
+  g_object_unref (invocation);
+
 out:
 
   return TRUE; /* handled */
